%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                   samsara.m
%                   Russell Luke
%  Department of MathSci, University of Delaware  
%                        and 
%  Inst. f. Num. u. Angew. Math., Universitaet Goettingen
%                    JUNE. 2, 2010  
%
%  This work was supported by NSF-DMS grant 0712796. 
%
%
% DESCRIPTION:  Main routine for the SAMSARA toolbox.
%
% INPUT:
%   input by driver: xold_vec, xnew_vec, fold, fnew, gradfold_vec,
%                              gradfnew_vec
%   input files (optional): OPTIONS_param, OPTIONS_tolerances,
%         OPTIONS_opt_choices  (this is a little awkward in Matlab, but 
%         is intended to match the design of the Fortran version) 
%
% OUTPUT:
%       xnew_vec, xold_vec, 
%
% USAGE: 
% 
% SUBROUTINE CALLS:
%      Hessian_product: invMSB2_product, invBFGS2_product, BFGS1_product,
%                       and MSB1_product
%      invHessian_product: MSB2_product, BFGS2_product, invBFGS1_product
%                       and invMSB1_product
%      Step_finder: 
%      Step_length
%      Scaler: delta_invBFGS' for BFGS1, else 'delta_MSB2'
%      History: 'cdSY_mat' for BFGS, `fdSY_mat' for Broyden implementations
%

function [xnew_vec,x0_vec,f0,gradf0_vec,stepsize]=...
    samsara(x0_vec, xnew_vec, f0, fnew,gradf0_vec, gradfnew_vec)

dim=length(x0_vec);
orthog_TOL=[];
step_TOL=[];
Maxit=[];
c=[];
beta=[];
gamma=[];
eta1=[];
eta2=[];
eta3=[];
eta4=[];
alpha_0=[];
maxmem=[];
lookback=[];
initial_step=[];
History=[];
QN_method=[];
update_conditions=[];
Step_finder=[];


%--------------------------------------------------------
%
%  Raed in options files.  By default, Matlab
%  reads in the options file in the local directory
%  before taking an identical filename in a more distant
%  branch of the path tree, so the user can effectively
%  override the samsara defaults by including/modifying
%  the OPTIONS_*.m files in the directoy from which the 
%  master/driver/calling routine is running.  
%
%-----------------------------------------------------------
OPTIONS_params
OPTIONS_tolerances
OPTIONS_opt_choices

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Tolerances and parameters:
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if(isempty(orthog_TOL))
  orthog_TOL=1e-6;
end
if(isempty(step_TOL))
  step_TOL  =1e-14;
end
if(isempty(Maxit))
    Maxit     = 500;
end
if(isempty(c))
    c     = .001;
end
if(isempty(beta))
    beta  = .99;
end
if(isempty(gamma))
   gamma = .5;
end
if(isempty(eta1))
    eta1 = .995;
end
if(isempty(eta2))
    eta2 = .8;
end
if(isempty(eta3))
    eta3 = .05;
end
if(isempty(eta4))
   eta4=.5*eta3;
end
if(isempty(alpha_0))
    alpha_0=1e-9;
end
if(isempty(maxmem))
    maxmem= 9;
end
if(isempty(lookback))
    lookback=0; 
    % parameter for nonmonotone methods.  The larger lookback, the more
                % steps the function evaluation is allowed to
                % increase.
end
if(isempty(initial_step))
    initial_step = 100*sqrt(gradf0_vec'*gradf0_vec);
end
if c> 0 && c< 1,
else
    disp('The slope modification parameter c in the backtracking subroutine is not in (0,1).')
    disp('Setting c to the default, c=0.001.')
    c     = .001;
end
if beta> 0 && beta< 1,
else
    disp('The slope modification parameter beta in the backtracking subroutine is not in (0,1).')
    disp('Setting beta to the default, beta=0.99.')
    beta  = .99;
end
if gamma>0 && gamma <1,
else
    disp('The backtracking parameter gamma is not in (0,1).')
    disp('Setting gamma to the default, gamma=0.5.')
    gamma = .5;
end
if eta1>eta2 && eta1 <1,
else
    disp('The trust region parameter eta1 is not in (eta2,1).')
    disp('Setting to the default, eta1=0.995.')
    eta1 = .995;
end
if eta2>eta3 && eta2 <eta1,
else
    disp('The trust region parameter eta2 is not in (eta3,eta1).')
    disp('Setting to the default, eta2=0.8.')
    eta2 = .8;
end
if eta3>0 && eta3 <eta2,
else
    disp('The trust region parameter eta3 is not in (0,eta2).')
    disp('Setting to the default, eta3=0.05.')
    eta3 = .05;
end
if alpha_0>1e-15,
elseif alpha_0<=1e-15,
    disp('WARNING: the regularization parameter is probably too small.')
    disp('We recommended choosing alpha_0>= 5e-12.')
else
    alpha_0=1e-9;
end
if step_TOL <= 0,
    error('The termination criteria step_TOL sent to the backtracking line search is not positive.')
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Set up appropriate matrix products for QN_method
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if(strcmp(History,'cdSY_mat')||strcmp(History,'fdSY_mat'))
else
    History='cdSY_mat';
end
if(strcmp(QN_method, 'Broyden1'))
    spsd=0;
    Hessian_product     = 'Broyden1_product'; 
    HessianT_product     = 'Broyden1_product'; 
    invHessian_product  = 'invBroyden1_product'; 
    Scaler = 'delta_MSB2';
    if(strcmp(update_conditions,'Exact'))
        Step_length = 'Exact';
    else
        Step_length = 'TR_adjust'; 
    end
    Step_finder = 'Dogleg_QN'; 
elseif(strcmp(QN_method, 'MSBroyden1'))
    spsd=0;
    Hessian_product     = 'MSB1_product';
    HessianT_product     = 'MSB1_product';
    invHessian_product  = 'invMSB1_product';
    Scaler = 'delta_MSB2';
    if(strcmp(update_conditions,'Exact'))
        Step_length = 'Exact';
    else
        Step_length = 'TR_adjust'; 
    end
    Step_finder='Dogleg_QN';
elseif(strcmp(QN_method, 'Broyden2'))
    spsd=0;
    Hessian_product     = 'invBroyden2_product'; 
    HessianT_product     = 'invBroyden2_product'; 
    invHessian_product  = 'Broyden2_product';
    Scaler = 'delta_MSB2';
    if(strcmp(update_conditions,'Exact'))
        Step_length = 'Exact';
    else
        Step_length = 'TR_adjust'; 
    end
    Step_finder = 'Dogleg_QN'; 
elseif(strcmp(QN_method, 'MSBroyden2'))
    spsd=0;
    Hessian_product     = 'invMSB2_product'; 
    HessianT_product     = 'invMSB2_product'; 
    invHessian_product  = 'MSB2_product';
    Scaler = 'delta_MSB2';
    if(strcmp(update_conditions,'Exact'))
        Step_length = 'Exact';
    else
        Step_length = 'TR_adjust'; 
    end
    Step_finder = 'Dogleg_QN'; 
elseif(strcmp(QN_method, 'BFGS1'))
    spsd=1;
    Hessian_product     = 'BFGS1_product'; 
    HessianT_product     = 'BFGS1_product'; 
    invHessian_product  = 'invBFGS1_product';
    Scaler = 'delta_invBFGS';
    if(strcmp(update_conditions,'Exact'))
        Step_length = 'Exact';
    else
        Step_length = 'TR_adjust'; 
    end
    if(isempty(Step_finder))
        Step_finder = 'Explicit_TR'; 
    end
    if(strcmp(History,'fdSY_mat'))
        disp('Finite difference history ordering not consistent with BFGS,')
        disp('   changing to conjugate direction history ordering, cdSY_mat.')
        History = 'cdSY_mat';
    end
elseif(strcmp(QN_method, 'MSBAI'))
    spsd=1;
    Hessian_product     = 'MSBA_product'; 
    HessianT_product     = 'MSBA_product'; 
    invHessian_product  = 'invMSBAI_product';
    Scaler = 'delta_MSB2';
    % Scaler = 'delta_invBFGS';
    if(strcmp(update_conditions,'Exact'))
        Step_length = 'Exact';
    else
        Step_length = 'TR_adjust'; 
    end
    if(isempty(Step_finder))
        Step_finder = 'Dogleg_QN'; 
    end
else
  disp('No such method.  Current options are:')
  disp('        Broyden1, MSBroyden1, Broyden2, MSBroyden2, BFGS1, BFGS2.')
  error('Stopping routine. ')
  % options under construction: 
  %     MSBFGS1, MSBFGS2, AvMSBroyden1, AvMSBroyden2, AvMSBFGS1, AvMSBFGS2
end
  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Read in history
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

fid=fopen('stephistory.dat','r');
if (fid<0)
    disp('History file nonexistent: initializating.')
    hist = 0;
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Initialization
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    iter           = 1;
    fcounter       = 1;
    trstep         = 0;
    mem            = -1;
    % the maxmem+1 column is a temporary storage area
    S_mat                   = [];
    Y_mat                   = [];
    stepsize_vec            = zeros(1,Maxit+1);
    ngradf0                 = sqrt(gradf0_vec'*gradf0_vec);
    if(isempty(gradfnew_vec))
        % automatically generate the first step
        d_vec               = -initial_step*gradf0_vec/ngradf0;
        xnew_vec            = x0_vec+d_vec;
        gradfnew_vec        = 999*ones(dim,1);
    else
        d_vec               = xnew_vec-x0_vec;
        fcounter       = fcounteer + 1;
    end
    stepsize                = feval('norm',d_vec);
    stepsize_vec(iter)      = stepsize;
    mu_vec                  = zeros(1,Maxit+1);
    lower_step_bound        = 0;
    alpha                   = alpha_0;%/delta;
    delta                   = 1;
    tr                      = stepsize;
    tmp_mat                 = zeros(dim,maxmem-1);
    x_mat(:,1:maxmem+1)     = [tmp_mat,x0_vec,xnew_vec];
    gradf_mat(:,1:maxmem+1) = [tmp_mat,gradf0_vec,gradfnew_vec];
    f_vec                   = zeros(1,Maxit+1);
    if isempty(f0)
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %   norm squared residual minimization if f0 and fnew empty:
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        [f0,gradf0_vec]=feval('Normsqresid',Scaler, History, Hessian_product, HessianT_product,...
            dim, mem,maxmem, spsd,...
            x_mat, gradf_mat, gradf0_vec, alpha_0);
        if(isempty(gradfnew_vec))
            % send back to calling driver to get gradfnew
            fnew=999;
        else
            [fnew,gradfnew_vec] = ...
                feval('Normsqresid',Scaler, History, Hessian_product, HessianT_product,...
                dim, mem,maxmem, spsd,...
                x_mat, gradf_mat, gradfnew_vec, alpha_0);
        end
    end
    if(isempty(fnew))
        % send back to calling driver to get gradfnew
        fnew=999;
    end
    f_vec(iter)             = f0;
    f_vec(iter+1)           = fnew;
    ngradf_vec              = zeros(1,Maxit+1);
    ngradf_vec(iter)        = ngradf0;
    ngradf_vec(iter+1)      = sqrt(gradfnew_vec'*gradfnew_vec);
    DDf                     = gradf_mat(:,maxmem)'*d_vec;
    cDDf                    = c*DDf;
    bDDf                    = beta*DDf;
else
    hist = 1;
    % read in all data, build all the components, and S_mat, and Y_mat
    stephist_mat  = fread(fid,[dim,2*maxmem+3],'real*8');
    fclose(fid);
    [fid,message] = fopen('state.dat','r');
    if (fid<0)
        disp(message)
    else
        state_vec = fread(fid,[1,4*(Maxit+1)+12],'real*8');
        fclose(fid);
    end
    iter                    = state_vec(4*(Maxit+1)+1);
    x_mat(:,1:maxmem+1)     = stephist_mat(:,1:maxmem+1);
    f_vec                   = state_vec(1:Maxit+1);
    d_vec                   = stephist_mat(:,2*maxmem+3);
    ngradf_vec              = state_vec(Maxit+2:2*(Maxit+1));
    stepsize_vec            = state_vec(2*Maxit+3:3*(Maxit+1));
    mu_vec                  = state_vec(3*Maxit+4:4*(Maxit+1));
    delta                   = state_vec(4*(Maxit+1)+2);
    alpha                   = state_vec(4*(Maxit+1)+3);
    tr                      = state_vec(4*(Maxit+1)+4);
    mem                     = state_vec(4*(Maxit+1)+5);
    DDf                     = state_vec(4*(Maxit+1)+6);
    cDDf                    = state_vec(4*(Maxit+1)+7);
    bDDf                    = state_vec(4*(Maxit+1)+8);
    lower_step_bound               = state_vec(4*(Maxit+1)+9);
    fcounter                = state_vec(4*(Maxit+1)+10)+1;
    trstep                  = state_vec(4*(Maxit+1)+11);
    gradf_mat(:,1:maxmem+1) = stephist_mat(:,maxmem+2:2*maxmem+2);
    if isempty(fnew)
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %   norm squared residual minimization if f0 and fnew empty
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        [fnew,gradfnew_vec] = ...
            feval('Normsqresid',Scaler, History, Hessian_product, HessianT_product,...
            dim, mem,maxmem, spsd,...
            x_mat, gradf_mat, gradfnew_vec, alpha_0);
    end
    f_vec(iter+1)           = fnew;
    gradf_mat(:,maxmem+1)   = gradfnew_vec;
    ngradf_vec(iter+1)      = sqrt(gradfnew_vec'*gradfnew_vec);
end
DDfnew = gradf_mat(:,maxmem+1)'*d_vec;
%-------------------------------------
%   end read in history
%-------------------------------------
if(hist==1)
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %   ADJUST STEP LENGTH/TRUST REGION and determine Update criteria
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    [tr,mu_vec,cDDf, lower_step_bound,d_vec, stepsize_vec, accept_step]=...
        feval('Step_length_update',History, Hessian_product, update_conditions,...
        mem,maxmem,iter,dim,...
        x_mat, f_vec,gradf_mat,d_vec, mu_vec,stepsize_vec,...
        tr,delta, alpha,eta1, eta2, eta3,cDDf,DDfnew, bDDf, orthog_TOL, lower_step_bound, gamma);

    %--------------------------------------
    % end  ADJUST STEP LENGTH/TRUST REGION and determine Update criteria
    %--------------------------------------

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %  Update the memory
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    [x_mat, f_vec, gradf_mat, ngradf_vec, stepsize_vec, mem] = ...
        feval('Memory_update',x_mat, f_vec, gradf_mat, ngradf_vec,...
        stepsize_vec, mem, maxmem, spsd, iter,accept_step);
    %--------------------------------------
    %  end update  memory
    %--------------------------------------

    muvec=mu_vec(iter);
    if(accept_step>-2)
        if(accept_step>=0) % If criteria satisfied, then update and increment iterates
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % reset lower_step_bound increment iter
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            lower_step_bound=0;
            iter=iter+1;
        end
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % choose new search direction (and length):
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        [delta, alpha, d_vec, stepsize_vec, trstep, DDf, cDDf, bDDf,mem] = ...
            feval('Step_generator',Step_finder,Scaler, History, Hessian_product, invHessian_product,...
            iter, dim, mem,maxmem, c, beta, gamma,...
            x_mat, gradf_mat, ngradf_vec, stepsize_vec,tr, trstep, alpha_0, verbose);
        %--------------------------------------
        %  end choose search direction and length
        %--------------------------------------
    end
end
% update the proposed step
x_mat(:,maxmem+1) = x_mat(:,maxmem)+d_vec;
if(verbose)
    display(['iteration: ',num2str(iter),'; trust region: ',num2str(tr),';  memory:' num2str(mem)])
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% write history, clean up and return
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
stephist_mat = [x_mat(:,1:maxmem+1),gradf_mat(:,1:maxmem+1),d_vec];
state_vec = [f_vec, ngradf_vec, stepsize_vec, mu_vec, iter, delta, ...
             alpha, tr, mem, DDf, cDDf, bDDf, lower_step_bound, fcounter, trstep, Maxit];
if(iter>=Maxit)
    display('Warning:   Maxit specified in OPTIONS_tolerances.m is smaller')
    display('  than the maximum iterations specified in the driver program.')
    display('  Increase the Maxit in OPTIONS_tolerances to at least that of')
    display('  the driver program.')
    error(' ')
end
[fid,message]=fopen('stephistory.dat','w');
if (fid<0)
    disp(message)
else
    stephist_mat=fwrite(fid,stephist_mat,'real*8');
    fclose(fid);
end

[fid,message]=fopen('state.dat','w');
if (fid<0)
    disp(message)
else
    state_vec=fwrite(fid,state_vec,'real*8');
    fclose(fid);
end
% new point:
xnew_vec = x_mat(:,maxmem+1);

% return old data, even though this is not really used
x0_vec    = x_mat(:,maxmem);
f0        = f_vec(:,iter);
gradf0_vec= gradf_mat(:,maxmem);
stepsize=stepsize_vec(iter);
% clear variables to avoid memory leaks
clear x_mat gradf_mat d_vec f_vec mu_vec stepsize_vec ngradf_vec
clear stephist_mat state_mat S_mat Y_mat

return