%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                  JWST_dvr.m
%                   Russell Luke
%  Department of MathSci, University of Delaware  
%                        and 
%  Inst. f. Num. u. Angew. Math., Universitaet Goettingen
%                    JUNE. 2, 2010  
%
%  This work was supported by NSF-DMS grant 0712796. 
%
%
% DESCRIPTION:  driver template for the SAMSARA reverse communication 
%               optimization toolbox.
%
% INPUT:  none...this is a script
%
% OUTPUT:
%       xnew_vec,  optimal solution
%       fnew, optimal value
%       diagnost_struct, data structure of diagnostic details, e.g. f_vec,
%                     ngrad_vec, stepsize_vec, iter and x_mat.
%
% USAGE: 
%        sample_dvr
% 
% SUBROUTINE CALLS:
%      samsara
%      Objective
%

clear;
% add samsara toolbox to directory path
addpath('../../');
% clear out old data files
!rm *.dat
!rm /Users/rluke/work/res/samsara/Matlab/*.dat
%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Function to minimize
%%%%%%%%%%%%%%%%%%%%%%%%%%%
Objective   = 'J_gradJ';  

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Initialization
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
im=sqrt(-1);
pi2=2*pi;
%========================= READ IN DATA============================

% Data parameters
res=128;   % For developing your algorithm you might want to 
           % start wih res=64 to make sure it works before getting
           % to res=512
noise=0;   % can add noise to the data -- not recommended.

Objective_parameters=feval('data_processor',res,noise);
dim = [res,res,Objective_parameters.diversity];
Objective_parameters.dim=dim;
%--------------------------
% initial guess from file
%--------------------------

% u_mat = Rbin('./data/u64.bin','r','l',[64,64],'float','complex');
% u_mat = Rbin('./data/u512.bin','r','l',[512,512],'float','complex');
temp_mat=MgenN(res,2*res,1e-3*pi2,0,20);
u_mat = temp_mat(:,1:res) + 1i*temp_mat(:,res+1:2*res);
u_mat = u_mat/norm(u_mat,'fro')*norm(Objective_parameters.k(:,:,1));
% load data/u_0.mat
% u_mat=u;
% clear u
u_mat=Resize(u_mat,res,res);
temp_mat = [real(u_mat),imag(u_mat)];
% % put image into vector form:
x0_vec=feval('reshape',temp_mat,2*res*res,1);
clear temp_mat


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Tolerances
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
ngrad_TOL = 2e-14;
step_TOL = .0001*ngrad_TOL;
Maxit=1500;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Clear old data
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!rm *.dat

%%%%%%%%%%%%%%%%%
% Main loop
%%%%%%%%%%%%%%%%%
iter=1;
xold_vec = x0_vec;
% SAMSARA needs two points:  the current point and a proposed point.  
% On the very first iteration the user can either generate the second 
% proposed point manually, or allow SAMSARA to automatically generate 
% the second point.  SAMSARA generates the second point by taking a scaling 
% of the steepest descent direction.  After the first iteration the current step is
% stored and updated in a file maintained by SAMSARA.  
[fold,gradfold_vec]=feval(Objective,Objective_parameters,xold_vec);
% fold=[];
% d_vec=-.1*gradfold_vec;
% xnew_vec=xold_vec+d_vec;
% stepsize=feval('norm',d_vec)
% [fnew,gradfnew_vec]=feval(Objective,xnew_vec);
% ngradfnew = sqrt(gradfnew_vec'*gradfnew_vec);
d_vec=[];
xnew_vec=[];
stepsize=999;
fnew=[];
gradfnew_vec = [];
ngradfnew=999;
while ((iter<Maxit)&&(ngradfnew>ngrad_TOL)&&(stepsize>step_TOL))
[xnew_vec,xold_vec,fold,gradfold_vec, stepsize]=...
        feval('samsara', xold_vec, xnew_vec, fold, fnew, gradfold_vec, gradfnew_vec);
    % new function evaluation:
    [fnew,gradfnew_vec] = feval(Objective,Objective_parameters,xnew_vec);
    ngradfnew = sqrt(gradfnew_vec'*gradfnew_vec);
end
if(stepsize<=step_TOL)
    disp('Algorithm stagnated:  stepsize tolerance violated.')
end
%%%%%%%%%%%%%%%%%
% end main loop
%%%%%%%%%%%%%%%%%
xopt_vec=xnew_vec;
fopt=fnew;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% prepare diagnostic data and return
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
[fid,message]=fopen('state.dat','r');
if (fid<0)
    disp(message)
else
    state_vec=fread(fid,[1,inf],'real*8');
    fclose(fid);
end
Maxit=state_vec(end);
f_vec = state_vec(1:Maxit+1);
ngradf_vec = state_vec(Maxit+2:2*(Maxit+1));
stepsize_vec = state_vec(2*Maxit+3:3*(Maxit+1));
mu_vec = state_vec(3*Maxit+4:4*(Maxit+1));
iter = state_vec(4*(Maxit+1)+1);
fcounter = state_vec(4*(Maxit+1)+10);
trstep = state_vec(4*(Maxit+1)+11);
JWST_viewer

display(['iterations: ',num2str(iter),'; function evaluations: ',num2str(fcounter),';  optimal value: ' num2str(fopt)])



